/*
 * Copyright (C) 2000-2024 the xine project
 *
 * This file is part of xine, a unix video player.
 *
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <pthread.h>

#include "common.h"
#include "tvset.h"
#include "videowin.h"
#include "actions.h"
#include "event.h"
#include "xine-toolkit/backend.h"
#include "xine-toolkit/combo.h"
#include "xine-toolkit/labelbutton.h"
#include "xine-toolkit/label.h"
#include "xine-toolkit/inputtext.h"
#include "xine-toolkit/intbox.h"

#include "frequencies.h"

#define WINDOW_WIDTH    500
#define WINDOW_HEIGHT   346


struct xui_tvset_s {
  gui_new_window_t      nw;

  xitk_widget_t        *input;
  xitk_widget_t        *system;
  xitk_widget_t        *chann;
  xitk_widget_t        *framerate;
  xitk_widget_t        *vidstd;

  xitk_widget_t        *close;
  xitk_widget_t        *update;

  const char          **system_entries;
  const char          **chann_entries;
  const char          **vidstd_entries;

  xitk_register_key_t   widget_key;

};

typedef uint64_t v4l2_std_id;

/* one bit for each */
#define V4L2_STD_PAL_B          ((v4l2_std_id)0x00000001)
#define V4L2_STD_PAL_B1         ((v4l2_std_id)0x00000002)
#define V4L2_STD_PAL_G          ((v4l2_std_id)0x00000004)
#define V4L2_STD_PAL_H          ((v4l2_std_id)0x00000008)
#define V4L2_STD_PAL_I          ((v4l2_std_id)0x00000010)
#define V4L2_STD_PAL_D          ((v4l2_std_id)0x00000020)
#define V4L2_STD_PAL_D1         ((v4l2_std_id)0x00000040)
#define V4L2_STD_PAL_K          ((v4l2_std_id)0x00000080)

#define V4L2_STD_PAL_M          ((v4l2_std_id)0x00000100)
#define V4L2_STD_PAL_N          ((v4l2_std_id)0x00000200)
#define V4L2_STD_PAL_Nc         ((v4l2_std_id)0x00000400)
#define V4L2_STD_PAL_60         ((v4l2_std_id)0x00000800)

#define V4L2_STD_NTSC_M         ((v4l2_std_id)0x00001000)
#define V4L2_STD_NTSC_M_JP      ((v4l2_std_id)0x00002000)

#define V4L2_STD_SECAM_B        ((v4l2_std_id)0x00010000)
#define V4L2_STD_SECAM_D        ((v4l2_std_id)0x00020000)
#define V4L2_STD_SECAM_G        ((v4l2_std_id)0x00040000)
#define V4L2_STD_SECAM_H        ((v4l2_std_id)0x00080000)
#define V4L2_STD_SECAM_K        ((v4l2_std_id)0x00100000)
#define V4L2_STD_SECAM_K1       ((v4l2_std_id)0x00200000)
#define V4L2_STD_SECAM_L        ((v4l2_std_id)0x00400000)

/* ATSC/HDTV */
#define V4L2_STD_ATSC_8_VSB     ((v4l2_std_id)0x01000000)
#define V4L2_STD_ATSC_16_VSB    ((v4l2_std_id)0x02000000)

static const struct {
  char     name[24];
  uint64_t std;
} std_list[] = {
  { "PAL-B",       V4L2_STD_PAL_B       },
  { "PAL-B1",      V4L2_STD_PAL_B1      },
  { "PAL-G",       V4L2_STD_PAL_G       },
  { "PAL-H",       V4L2_STD_PAL_H       },
  { "PAL-I",       V4L2_STD_PAL_I       },
  { "PAL-D",       V4L2_STD_PAL_D       },
  { "PAL-D1",      V4L2_STD_PAL_D1      },
  { "PAL-K",       V4L2_STD_PAL_K       },
  { "PAL-M",       V4L2_STD_PAL_M       },
  { "PAL-N",       V4L2_STD_PAL_N       },
  { "PAL-Nc",      V4L2_STD_PAL_Nc      },
  { "PAL-60",      V4L2_STD_PAL_60      },
  { "NTSC-M",      V4L2_STD_NTSC_M      },
  { "NTSC-M-JP",   V4L2_STD_NTSC_M_JP   },
  { "SECAM-B",     V4L2_STD_SECAM_B     },
  { "SECAM-D",     V4L2_STD_SECAM_D     },
  { "SECAM-G",     V4L2_STD_SECAM_G     },
  { "SECAM-H",     V4L2_STD_SECAM_H     },
  { "SECAM-K",     V4L2_STD_SECAM_K     },
  { "SECAM-K1",    V4L2_STD_SECAM_K1    },
  { "SECAM-L",     V4L2_STD_SECAM_L     },
  { "ATSC-8-VSB",  V4L2_STD_ATSC_8_VSB  },
  { "ATSC-16-VSB", V4L2_STD_ATSC_16_VSB }
};


static void tvset_update (xitk_widget_t *w, void *data, int state) {
  xui_tvset_t *tvset = data;
  xine_event_t          xine_event;
  xine_set_v4l2_data_t *ev_data;
  int current_system, current_chan, current_std;

  (void)w;
  (void)state;
  current_system = xitk_combo_get_current_selected (tvset->system);
  current_chan   = xitk_combo_get_current_selected (tvset->chann);
  current_std    = xitk_combo_get_current_selected (tvset->vidstd);

  if (current_system < 0 || (size_t)current_system >= sizeof(chanlists) / sizeof(chanlists[0])) {
    current_system = 0;
  }
  if (current_chan < 0 || current_chan >= chanlists[current_system].count) {
    current_chan = 0;
  }
  if (current_std < 0 || (size_t)current_std >= sizeof(std_list) / sizeof(std_list[0])) {
    current_std = 0;
  }

  ev_data = (xine_set_v4l2_data_t *)malloc(sizeof(xine_set_v4l2_data_t));
  if (!ev_data)
    return;

  ev_data->input     = xitk_intbox_get_value (tvset->input);
  ev_data->frequency = (chanlists[current_system].list[current_chan].freq * 16) / 1000;

  ev_data->standard_id = std_list[current_std].std;

  xine_event.type        = XINE_EVENT_SET_V4L2;
  xine_event.data_length = sizeof(xine_set_v4l2_data_t);
  xine_event.data        = ev_data;
  xine_event.stream      = tvset->nw.gui->stream;
  xitk_gettime_tv (&xine_event.tv);

  xine_event_send (tvset->nw.gui->stream, &xine_event);
}

static void tvset_exit (xitk_widget_t *w, void *data, int state) {
  xui_tvset_t *tvset = data;
  window_info_t wi;

  (void)w;
  (void)state;
  if (!tvset)
    return;

  if ((xitk_get_window_info (tvset->nw.gui->xitk, tvset->widget_key, &wi))) {
    config_update_num (tvset->nw.gui->xine, "gui.tvset_x", wi.x);
    config_update_num (tvset->nw.gui->xine, "gui.tvset_y", wi.y);
  }

  tvset->nw.gui->tvset = NULL;

  xitk_unregister_event_handler (tvset->nw.gui->xitk, &tvset->widget_key);
  xitk_window_destroy_window (tvset->nw.xwin);
  tvset->nw.xwin = NULL;
  /* xitk_dlist_init (&tvset->nw.wl->list); */

  free (tvset->system_entries);
  tvset->system_entries = NULL;
  free (tvset->chann_entries);
  tvset->chann_entries = NULL;
  free (tvset->vidstd_entries);
  tvset->vidstd_entries = NULL;

  video_window_set_input_focus (tvset->nw.gui->vwin);

  free (tvset);
}

static int tvset_event (void *data, const xitk_be_event_t *e) {
  xui_tvset_t *tvset = data;

  if (((e->type == XITK_EV_KEY_DOWN) && (e->utf8[0] == XITK_CTRL_KEY_PREFIX) && (e->utf8[1] == XITK_KEY_ESCAPE))
    || (e->type == XITK_EV_DEL_WIN)) {
    tvset_exit (NULL, tvset, 0);
    return 1;
  }
  return gui_handle_be_event (tvset->nw.gui, e);
}

static int update_chann_entries (xui_tvset_t *tvset, int system_entry) {
  int               i;
  const struct CHANLIST *list = chanlists[system_entry].list;
  int               len  = chanlists[system_entry].count;

  free (tvset->chann_entries);

  tvset->chann_entries = (const char **) calloc((len+1), sizeof(const char *));

  for(i = 0; i < len; i++)
    tvset->chann_entries[i] = list[i].name;

  tvset->chann_entries[i] = NULL;
  return len;
}

static void system_combo_select (xitk_widget_t *w, void *data, int select) {
  xui_tvset_t *tvset = data;
  int len;

  (void)w;
  len = update_chann_entries (tvset, select);

  if (tvset->chann) {
    xitk_combo_update_list (tvset->chann, tvset->chann_entries, len);
    xitk_combo_set_select (tvset->chann, 0);
  }
}

void tvset_main (xitk_widget_t *mode, void *data) {
  gGui_t *gui = data;
  xui_tvset_t  *tvset;
  size_t        i;
  int           x, y, w;

  if (!gui)
    return;

  tvset = gui->tvset;
  if (mode == XUI_W_OFF) {
    if (!tvset)
      return;
    tvset_exit (NULL, tvset, 0);
    return;
  } else if (mode == XUI_W_ON) {
    if (tvset) {
      gui_raise_window (gui, tvset->nw.xwin, 1, 0);
      xitk_window_set_input_focus (tvset->nw.xwin);
      return;
    }
  } else { /* toggle */
    if (tvset) {
      tvset_exit (NULL, tvset, 0);
      return;
    }
  }

  tvset = calloc (1, sizeof (*tvset));
  if (!tvset)
    return;

  tvset->nw.gui = gui;

  /* Create window */
  tvset->nw.title = _("TV Analog Video Parameters");
  tvset->nw.id = "tvset";
  if (xitk_init_NULL ()) {
    tvset->nw.skin = NULL;
    tvset->nw.wfskin = NULL;
    tvset->nw.adjust = NULL;
  }
  tvset->nw.wr.x = 80;
  tvset->nw.wr.y = 80;
  tvset->nw.wr.width = WINDOW_WIDTH;
  tvset->nw.wr.height = WINDOW_HEIGHT;
  if (gui_window_new (&tvset->nw) < 0) {
    free (tvset);
    return;
  }

  x = 15;
  y = 34 - 6;
  xitk_image_draw_frame (tvset->nw.bg, _("General"), btnfontname, x, y, WINDOW_WIDTH - 30, ((20 + 22) + 5 + 2) + 15, XITK_DRAW_OUTTER);

  /* First Line */
  x = 20;
  y += 15;
  w = 139;
  xitk_image_draw_frame (tvset->nw.bg, _("Input: "), lfontname, x, y, w, 20 + 22, XITK_DRAW_INNER);

  {
    xitk_intbox_widget_t ib = {
      .nw = { .wl = tvset->nw.wl, .userdata = tvset, .add_state = XITK_WIDGET_STATE_ENABLE | XITK_WIDGET_STATE_VISIBLE },
      ib.fmt    = INTBOX_FMT_DECIMAL,
      ib.min    = 0,
      ib.max    = 0,
      ib.value  = 4,
      ib.step   = 1
    };
    tvset->input = xitk_noskin_intbox_create (&ib, x + 10, y + 15, w - 20 + 1, 20);
  }

  {
    static const size_t chanlists_count = sizeof(chanlists)/sizeof(chanlists[0]);
    tvset->system_entries = (const char **)calloc ((chanlists_count + 1), sizeof (const char *));

    for(i = 0; i < chanlists_count; i++)
      tvset->system_entries[i] = chanlists[i].name;
    tvset->system_entries[i] = NULL;
  }

  x += w + 5;
  w = 155;
  xitk_image_draw_frame (tvset->nw.bg, _("Broadcast System: "), lfontname, x, y, w, 20 + 22, XITK_DRAW_INNER);

  {
    xitk_combo_widget_t cmb = {
      .nw = { .wl = tvset->nw.wl, .userdata = tvset, .add_state = XITK_WIDGET_STATE_ENABLE | XITK_WIDGET_STATE_VISIBLE },
      .layer_above       = gui_is_layer_above (gui),
      .parent_wkey       = &tvset->widget_key
    };
    xitk_inputtext_widget_t inp = {
        .nw = { .wl = tvset->nw.wl, .userdata = tvset, .add_state = XITK_WIDGET_STATE_ENABLE | XITK_WIDGET_STATE_VISIBLE },
    };

    cmb.entries  = tvset->system_entries;
    cmb.callback = system_combo_select;
    tvset->system = xitk_noskin_combo_create (&cmb, x + 10, y + 15, w - 20 + 1, 18);
    xitk_set_widget_pos (tvset->system, x + 10, y + 15 + (20 - xitk_get_widget_height (tvset->system)) / 2);

    xitk_combo_set_select (tvset->system, 0);
    update_chann_entries (tvset, 0);

    x += w + 5;
    w = 155;
    xitk_image_draw_frame (tvset->nw.bg, _("Channel: "), lfontname, x, y, w, 20 + 22, XITK_DRAW_INNER);

    cmb.entries  = tvset->chann_entries;
    cmb.callback = NULL;
    tvset->chann = xitk_noskin_combo_create (&cmb, x + 10, y + 15, w - 20 + 1, 18);
    xitk_set_widget_pos (tvset->chann, x + 10, y + 15 + (20 - xitk_get_widget_height (tvset->chann)) / 2);

    x = 15;
    y += ((20 + 22) + 5 + 2) + 3;
    xitk_image_draw_frame (tvset->nw.bg, _("Standard"), btnfontname, x, y, WINDOW_WIDTH - 30, ((20 + 22) + 5 + 2) + 15, XITK_DRAW_OUTTER);

    x = 20;
    y += 15;
    w = 139;
    xitk_image_draw_frame (tvset->nw.bg, _("Frame Rate: "), lfontname, x, y, w, 20 + 22, XITK_DRAW_INNER);

    inp.max_length = 20;
    tvset->framerate = xitk_noskin_inputtext_create (&inp,
      x + 10, y + 15, w - 20 + 1, 20, XITK_NOSKIN_TEXT_NORM, XITK_NOSKIN_TEXT_NORM, fontname);

    tvset->vidstd_entries = (const char **) malloc(sizeof(const char *) *
      (sizeof(std_list)/sizeof(std_list[0])+1));

    for(i = 0; i < (sizeof(std_list)/sizeof(std_list[0])); i++)
      tvset->vidstd_entries[i] = std_list[i].name;
    tvset->vidstd_entries[i] = NULL;

    x += w + 5;
    w = 155;
    xitk_image_draw_frame (tvset->nw.bg, _("Analog Standard: "), lfontname, x, y, w, 20 + 22, XITK_DRAW_INNER);

    cmb.entries  = tvset->vidstd_entries;
    cmb.callback = NULL;
    tvset->vidstd = xitk_noskin_combo_create (&cmb, x + 10, y + 15, w - 20 + 1, 18);

    x = 15;
    y += ((20 + 22) + 5 + 2) + 3;
    xitk_image_draw_frame (tvset->nw.bg, _("Frame Size"), btnfontname, x, y, WINDOW_WIDTH - 30, ((20 + 22) + 5 + 2) + 15, XITK_DRAW_OUTTER);

    x = 20;
    y += 15;
    w = 139;

    x = 15;
    y += ((20 + 22) + 5 + 2) + 3;
    xitk_image_draw_frame (tvset->nw.bg, _("MPEG2"), btnfontname, x, y, WINDOW_WIDTH - 30, ((20 + 22) + 5 + 2) + 15, XITK_DRAW_OUTTER);

    x = 20;
    y += 15;
    w = 139;

    /*  */
    y = WINDOW_HEIGHT - (23 + 15);
    x = 15;
  }

  {
    xitk_labelbutton_widget_t lb = {
      .nw = { .wl = tvset->nw.wl, .userdata = tvset, .add_state = XITK_WIDGET_STATE_ENABLE | XITK_WIDGET_STATE_VISIBLE },
      .button_type = CLICK_BUTTON,
      .align       = ALIGN_CENTER
    };

    lb.label    = _("Update");
    lb.callback = tvset_update;
    tvset->update = xitk_noskin_labelbutton_create (&lb,
      x, y, 100, 23, XITK_NOSKIN_TEXT_NORM, XITK_NOSKIN_TEXT_NORM, XITK_NOSKIN_TEXT_INV, btnfontname);

    x = WINDOW_WIDTH - (100 + 15);

    lb.label    = _("Close");
    lb.callback = tvset_exit;
    xitk_noskin_labelbutton_create (&lb,
      x, y, 100, 23, XITK_NOSKIN_TEXT_NORM, XITK_NOSKIN_TEXT_NORM, XITK_NOSKIN_TEXT_INV, btnfontname);
  }

  xitk_window_set_background_image (tvset->nw.xwin, tvset->nw.bg);

  tvset->widget_key = xitk_be_register_event_handler ("tvset", tvset->nw.xwin, tvset_event, tvset, NULL, NULL);
  gui->tvset = tvset;

  gui_raise_window (tvset->nw.gui, tvset->nw.xwin, 1, 0);

  xitk_window_set_input_focus (tvset->nw.xwin);
}
